"""@anchor pydoc:grizzly.tasks.wait_between Wait Between
This task sets the wait time between tasks in a scenario.

The default is to wait `0` seconds between each task.

This is useful in a scenario with many tasks that should have some wait time between them, but there are a group
of tasks (e.g. Transform, Date or Log Messages) that should execute as fast as possible.

If `max_time` is not provided, the wait between tasks is constant `min_time`. If both are provided there will be a
random wait between (and including) `min_time` and `max_time` between tasks.

## Step implementations

* {@pylink grizzly.steps.scenario.tasks.wait_between.step_task_wait_between_constant}

* {@pylink grizzly.steps.scenario.tasks.wait_between.step_task_wait_between_random}

## Statistics

This task does not have any request statistics entries.

## Arguments

* `min_time` _float_ - minimum time to wait

* `max_time` _float_ (optional) - maximum time to wait
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from locust import between, constant

from . import GrizzlyTask, grizzlytask

if TYPE_CHECKING:  # pragma: no cover
    from grizzly.scenarios import GrizzlyScenario


class WaitBetweenTask(GrizzlyTask):
    min_time: float
    max_time: Optional[float]

    def __init__(self, min_time: float, max_time: Optional[float] = None) -> None:
        super().__init__()

        self.min_time = min_time
        self.max_time = max_time

    def __call__(self) -> grizzlytask:
        @grizzlytask
        def task(parent: GrizzlyScenario) -> Any:
            wait_time = constant(self.min_time) if self.max_time is None else between(self.min_time, self.max_time)

            bound_wait_time = wait_time.__get__(parent.user, parent.user.__class__)

            parent.user.wait_time = bound_wait_time

        return task
